"""setup an Idefix problem"""
from __future__ import annotations

import os
import re
import shutil
import subprocess
from argparse import ArgumentParser
from typing import NoReturn

from packaging.version import Version

from idefix_cli._commons import get_idefix_version
from idefix_cli._commons import get_user_conf_requirement
from idefix_cli._commons import get_user_config_file
from idefix_cli._commons import print_err
from idefix_cli._commons import print_warning
from idefix_cli._commons import requires_idefix

VERSION_REGEXP = re.compile(r"\d+\.\d+\.\d+")

CMAKE_MIN_VERSIONS: dict[str, Version] = {
    "cmake": Version("3.16.0"),
    "idefix": Version("0.9.0"),
}


class IdefixEnvError(OSError):
    pass


def validate_cmake_support() -> None:
    msg = f"cmake is required from {get_user_config_file()}, but "
    errors: list[str] = []
    warnings: list[str] = []
    idefix_ver = get_idefix_version()

    if idefix_ver is None or idefix_ver < CMAKE_MIN_VERSIONS["idefix"]:
        # try a simpler heuristic to allow usage in between 0.8.x and 0.9.0
        if os.path.isfile(os.path.join(os.environ["IDEFIX_DIR"], "CMakeLists.txt")):
            warnings.append(
                "looks like your version of Idefix predates 0.9.0, but already has "
                "cmake support. Results may be unstable."
            )
        else:
            errors.append(
                f"cmake setup requires idefix {CMAKE_MIN_VERSIONS['idefix']} or newer, "
                f"found {idefix_ver}"
            )

    if shutil.which("cmake") is None:
        errors.append("couldn't find cmake executable")
    else:
        cmake_ver: str = subprocess.run(
            ["cmake", "--version"], capture_output=True
        ).stdout.decode()
        if (match := re.search(VERSION_REGEXP, cmake_ver)) is None:
            errors.append("couldn't parse result from `cmake --version`")

        elif (cmake_ver := Version(match.group())) < CMAKE_MIN_VERSIONS["cmake"]:
            errors.append(
                f"cmake setup requires cmake {CMAKE_MIN_VERSIONS['cmake']} or newer, "
                f"found {cmake_ver}"
            )

    for msg in warnings:
        print_warning(msg)

    if len(errors) == 0:
        return

    if len(errors) == 1:
        msg += errors[0]
    else:
        msg += "\n- " + "\n- ".join(errors)

    raise IdefixEnvError(msg)


def has_minimal_cmake_support() -> bool:
    try:
        validate_cmake_support()
    except IdefixEnvError:
        return False
    else:
        return True


def is_cmake_required() -> bool:
    req = get_user_conf_requirement("idefix_cli", "conf_system")
    return req is not None and req == "cmake"


def is_python_required() -> bool:
    req = get_user_conf_requirement("idefix_cli", "conf_system")
    return req is not None and req == "python"


@requires_idefix()
def has_python_support() -> bool:
    return os.path.isfile(os.path.join(os.environ["IDEFIX_DIR"], "configure.py"))


@requires_idefix()
def validate_python_support() -> None:
    if has_python_support():
        return

    msg = "Running a version of Idefix that doesn't provide $IDEFIX_DIR/configure.py . "
    if is_python_required():
        msg += f"This configuration system was required from {get_user_config_file()}"
    raise IdefixEnvError(msg)


def get_valid_conf_system() -> str:
    cmake_is_valid = has_minimal_cmake_support()
    python_is_valid = has_python_support()
    if cmake_is_valid:
        return "cmake"
    elif python_is_valid:
        return "python"
    else:
        raise IdefixEnvError(
            "Could not determine a working configuration system. "
            "Most likely, your version of Idefix requires CMake, "
            "which is currently not installed. "
            "Please consult Idefix's documentation, "
            "or try to update idefix-cli if it didn't help."
        )


def substitute_cmake_flags(args: list[str]) -> list[str]:
    subs: dict[str, str] = {
        "-mhd": "-DIdefix_MHD=ON",
        "-mpi": "-DIdefix_MPI=ON",
        "-openmp": "-DKokkos_ENABLE_OPENMP=ON",
        "-gpu": "-DKokkos_ENABLE_CUDA=ON",
    }
    return [subs.get(_) or _ for _ in args]


def substitute_cmake_archs(args: list[str]) -> list[str]:
    if not (
        "-arch" in args or any(re.match(r"-D\s?Kokkos_ARCH_\w+=ON", _) for _ in args)
    ):
        for option_name in ("CPU", "GPU"):
            arch_req = get_user_conf_requirement("compilation", option_name)
            if arch_req is None:
                continue
            args.extend(["-arch", arch_req])

    parser = ArgumentParser()
    parser.add_argument("-arch", nargs="+", required=False)
    _args, unknown_args = parser.parse_known_args(args)

    if _args.arch is None:
        return args

    return unknown_args + [f"-DKokkos_ARCH_{_.upper()}=ON" for _ in _args.arch]


def substitute_cmake_cxx(args: list[str]) -> list[str]:
    if not (
        "-cxx" in args or any(re.match(r"-DCMAKE_CXX_COMPILER=\w+", _) for _ in args)
    ):
        compiler_req = get_user_conf_requirement("compilation", "compiler")
        if compiler_req is not None:
            args.extend(["-cxx", compiler_req])

    parser = ArgumentParser()
    parser.add_argument("-cxx")
    _args, unknown_args = parser.parse_known_args(args)

    if _args.cxx is None:
        return args

    return unknown_args + [f"-DCMAKE_CXX_COMPILER={_args.cxx}"]


def substitute_cmake_args(args: list[str]) -> list[str]:
    # compatibility layer to enable configure.py's arguments with cmake
    args = substitute_cmake_flags(args)
    args = substitute_cmake_archs(args)
    args = substitute_cmake_cxx(args)
    return args


parser_kwargs = dict(
    add_help=False
)  # because it's a wrapper, we want to pass down even the "--help" flag


def add_arguments(parser) -> None:
    parser.add_argument(
        "-i",
        "--interactive",
        dest="interactive",
        action="store_true",
        help="Use ccmake over cmake " "(no effect with python configuration system)",
    )


@requires_idefix()
def command(*args: str, interactive: bool) -> int | NoReturn:
    python_cmd = ["python3", os.path.join(os.environ["IDEFIX_DIR"], "configure.py")]
    cmake_cmd = ["ccmake" if interactive else "cmake", os.environ["IDEFIX_DIR"]]
    system_req = get_user_conf_requirement("idefix_cli", "conf_system")

    if system_req is None:
        try:
            system_req = get_valid_conf_system()
        except IdefixEnvError as exc:
            print_err(exc)
            return 1
    else:
        try:
            validate_selected_system = {
                "cmake": validate_cmake_support,
                "python": validate_python_support,
            }[system_req]
        except KeyError:
            print_err(
                f"Got unknown value conf_system={system_req!r} "
                f"from {get_user_config_file()}, "
                "expected 'cmake' or 'python'"
            )
            return 1

        try:
            validate_selected_system()
        except IdefixEnvError as exc:
            print_err(exc)
            return 1

    clargs = list(args)
    if system_req == "cmake":
        cmd = cmake_cmd
        clargs = substitute_cmake_args(clargs)
    else:
        assert system_req == "python"
        cmd = python_cmd

    cmd.extend(clargs)
    print(f"INFO: running '{' '.join(cmd)}'")
    os.execvp(cmd[0], cmd)



"""clone a problem directory"""
from __future__ import annotations

import os
import shutil
from glob import glob
from tempfile import TemporaryDirectory

from rich import print

from idefix_cli._commons import files_from_patterns
from idefix_cli._commons import print_err

minimal_target = frozenset(
    (
        "idefix.ini",
        "definitions.hpp",
        "setup.cpp",
    )
)


def add_arguments(parser) -> None:
    parser.add_argument(
        "source", default=".", help="the problem directory to be cloned."
    )
    parser.add_argument(
        "dest",
        help="destination directory (must not exist, unless the --force flag is used.).",
    )
    parser.add_argument(
        "--shallow",
        action="store_true",
        help="build symlinks instead of actual copies (not implemented yet).",
    )
    parser.add_argument(
        "--extra",
        nargs="+",
        help="a list of additional file names (or patterns) that should be included in the clone.",
    )


def command(
    source,
    dest,
    shallow: bool = False,
    extra: list[str] | None = None,
) -> int:
    if not os.path.isdir(source):
        print_err(f"Error: source directory not found {source}")
        return 1
    if not os.listdir(source):
        print_err(f"Error: {source} appears to be empty.")
        return 1
    if os.path.exists(dest):
        print_err(f"Error: destination directory exists {dest}")
        return 1

    if extra is None:
        extra = []

    files_to_generate = files_from_patterns(source, *minimal_target, *extra)
    if not files_to_generate:
        print_err(f"Error: did not find any file to copy from {source}")
        return 1

    with TemporaryDirectory(dir=os.path.dirname(dest)) as tmpdir:
        # using a context manager to guarantee atomicity:
        # either it's a full success or we don't copy anything
        # The temporary directory is created next to the final destination
        # so we can safely use os.replace (atomic) without worrying about
        # possibly sparse filesystems.
        for file in files_from_patterns(source, *minimal_target, *extra):
            fdest = os.path.join(tmpdir, os.path.basename(file))
            if shallow:
                os.symlink(file, fdest)
            else:
                shutil.copy(file, fdest)
        os.replace(tmpdir, dest)

    output_files = list(glob(os.path.join(dest, "*")))
    if shallow:
        objs = "symlinks"
        lencol1 = max(len(_) for _ in output_files)
        lines = [f"{_.ljust(lencol1 + 1)} -> {os.readlink(_)}" for _ in output_files]
    else:
        objs = "files"
        lines = output_files
    output_files = "\n".join(list(glob(os.path.join(dest, "*"))))
    print(f"Created the following {objs}\n" + "\n".join(lines))
    return 0



from __future__ import annotations

import contextlib
import os
import platform
import re
import sys
from configparser import ConfigParser
from datetime import datetime
from functools import wraps
from getpass import getuser
from glob import glob
from itertools import chain
from multiprocessing import cpu_count
from socket import gethostname
from subprocess import CalledProcessError
from subprocess import check_call
from time import ctime
from typing import Any
from typing import Callable
from typing import cast
from typing import TypeVar
from typing import Union

from packaging.version import Version
from rich.console import Console

# workaround mypy not being confortable around decorator preserving signatures
# adapted from
# https://github.com/python/mypy/issues/1551#issuecomment-253978622
TFun = TypeVar("TFun", bound=Callable[..., Any])

VERSION_STR = r"\d+\.\d+\.\d+"
VERSION_REGEXP = re.compile(VERSION_STR)
# This is how sections are formatted in the changelog,
# e.g., '## [0.8.1] - 2021-06-24'
VERSECT_REGEXP = re.compile(fr"## \[{VERSION_STR}\] - \d\d\d\d-\d\d-\d\d\s*\n")

if platform.system().lower().startswith("win"):
    # Windows
    env_var = "APPDATA"
    default_usr_dir = "AppData"
else:
    # POSIX
    env_var = "XDG_CONFIG_HOME"
    default_usr_dir = ".config"

XDG_CONFIG_HOME = os.environ.get(
    env_var,
    os.path.join(os.path.expanduser("~"), default_usr_dir),
)
del env_var, default_usr_dir


class requires_idefix:
    def __call__(self, f: TFun) -> TFun:
        @wraps(f)
        def wrapper(*args, **kwargs) -> Any:
            if (IDEFIX_DIR := os.getenv("IDEFIX_DIR")) is None:
                print_err(
                    "this functionality requires $IDEFIX_DIR to be defined.",
                )
                return 10
            elif not os.path.isdir(IDEFIX_DIR):
                print_err(
                    f"env variable $IDEFIX_DIR isn't properly defined: {IDEFIX_DIR} is not a directory.",
                )
                return 20
            return f(*args, **kwargs)

        return cast(TFun, wrapper)


# setup a very large console to prevent rich from wrapping long error messages
# since it usually has the side effect of truncating file names.
# This workaround also helps keeping error messages reproducible in CI.

ErrorMessage = Union[str, Exception]


def print_err(message: ErrorMessage) -> None:
    err_console = Console(width=500, file=sys.stderr)
    err_console.print(f"[bold white on red]ERROR[/] {message}")


def print_warning(message: ErrorMessage) -> None:
    err_console = Console(width=500, file=sys.stderr)
    err_console.print(f"[red]WARNING[/] {message}")


@contextlib.contextmanager
def pushd(new_dir):
    previous_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(previous_dir)


@requires_idefix()
def _make(directory) -> int:
    ncpus = str(min(8, cpu_count() // 2))
    try:
        with pushd(directory):
            return check_call(["make", "-j", ncpus])
    except CalledProcessError as exc:
        print_err("failed to compile idefix")
        return exc.returncode


def files_from_patterns(source, *patterns) -> list[str]:
    raw = sorted(chain.from_iterable(glob(os.path.join(source, p)) for p in patterns))
    # it is important to append os.path.sep to directory names so
    # it's clear that they are not files when listed with idfx clean --dry
    return [_ + os.path.sep if os.path.isdir(_) else _ for _ in raw]


@requires_idefix()
def get_git_data() -> dict[str, str]:
    data = {
        "user": getuser(),
        "host": gethostname(),
        "date": ctime(datetime.now().timestamp()),
    }
    try:
        # this import may fail in envs where the git executable is not present,
        # so we'll avoid keeping it at the top level to minimize breakage
        import git
    except ImportError as exc:
        print_warning(f"failed to load gitpython (got 'ImportError: {exc}')")
    else:
        repo = git.Repo(os.environ["IDEFIX_DIR"])
        data = {"sha": repo.head.object.hexsha, **data}
    if (version := get_idefix_version()) is None:
        version_str = "unknown version"
    else:
        version_str = str(version)
    data = {"latest ancestor version": version_str, **data}
    return data


@requires_idefix()
def get_idefix_version() -> Version | None:
    # We rely on parsing the CHANGELOG file to determine the most recent release at
    # any given point. This is more reliable than checking for the closest ancestor
    # in git tags because the development branch usually doesn't decend from releases.
    # Another reason why this seems reasonable is that tarball releases are supposed
    # to ship a CHANGELOG file as well.
    changelog = os.path.join(os.environ["IDEFIX_DIR"], "CHANGELOG.md")
    if not os.path.isfile(changelog):
        # this is inevitable if the user is checked in a version that predates
        # the introduction of a changelog (Idefix v0.7.0)
        return None

    with open(changelog) as fh:
        for line in fh:
            if not re.match(VERSECT_REGEXP, line):
                continue
            match = re.search(VERSION_REGEXP, line)
            assert match is not None
            return Version(match.group())

    raise RuntimeError(
        "Something went wrong while trying to determine Idefix's version."
    )


def get_user_config_file() -> str | None:
    for parent_dir in [".", XDG_CONFIG_HOME]:
        if os.path.isfile(conf_file := os.path.join(parent_dir, "idefix.cfg")):
            return os.path.abspath(conf_file)
    return None


def get_user_configuration() -> ConfigParser | None:
    if (conf_file := get_user_config_file()) is None:
        return None

    cf = ConfigParser()
    cf.read(conf_file)
    return cf


def get_user_conf_requirement(section_name: str, option_name: str, /) -> str | None:
    if (usr_conf := get_user_configuration()) is None:
        return None

    return usr_conf.get(section_name, option_name, fallback=None)



"""clean up generated files"""
from __future__ import annotations

import os
import subprocess
from pathlib import Path
from shutil import rmtree
from shutil import which

from rich.prompt import Confirm

from idefix_cli._commons import files_from_patterns
from idefix_cli._commons import pushd

# bpatterns are those targeted by `make clean`, which is equivalent to
# rm -f *.o *.cuda *.host
bpatterns = frozenset(("*.o", "*.cuda", "*.host"))

# additional generated files
kokkos_files = frozenset(
    (
        "KokkosCore_config.h",
        "KokkosCore_config.tmp",
        "KokkosCore_Config*.tmp",
        "KokkosCore_Config*.hpp",
        "libkokkos.a",
    )
)

cmake_files = frozenset(("CMakeCache.txt", "cmake_install.cmake", "build"))

# only cleared if `--all` flag is passed
gpatterns = frozenset(("Makefile", "idefix"))

GENERATED_DIRS = frozenset(("CMakeFiles",))


def add_arguments(parser) -> None:
    parser.add_argument(
        "directory", nargs="?", default=".", help="the target directory to clean"
    )
    parser.add_argument(
        "--all",
        dest="clean_all",
        action="store_true",
        help="also clean generated Makefiles, idefix excutable files",
    )
    parser.add_argument(
        "--dry-run",
        "--dry",
        dest="dry",
        action="store_true",
        help="skip prompt, exit without cleaning",
    )


def command(directory, clean_all: bool = False, dry: bool = False) -> int:
    with pushd(directory):
        patterns = bpatterns | kokkos_files | cmake_files | GENERATED_DIRS
        if clean_all:
            patterns |= gpatterns

        # Guarantee that git indexed files are never cleaned
        if which("git") is not None:
            # we don't check the result of the process, meaning we trust git not to output
            # anything to stdout in case the current dir isn't part of a git repo
            git_files = set(
                subprocess.run(["git", "ls-files"], capture_output=True)
                .stdout.decode()
                .strip()
                .split("\n")
            )
        else:
            git_files = set()

        patterns -= git_files

        targets = files_from_patterns(Path.cwd(), *patterns)

        if not targets:
            print("Nothing to remove.")
            return 0

        print("The following files and directories can be removed")
        print("\n".join(targets))

        if not dry and Confirm.ask("\nPerform cleaning ? (y/[n])"):
            for t in targets:
                if os.path.isdir(t):
                    rmtree(t)
                else:
                    os.remove(t)

    return 0



"""run an Idefix problem"""
from __future__ import annotations

from pathlib import Path
from subprocess import call
from tempfile import NamedTemporaryFile

import inifix

from idefix_cli._commons import _make
from idefix_cli._commons import print_err
from idefix_cli._commons import pushd


def add_arguments(parser) -> None:

    parser.add_argument("directory", nargs="?", default=".", help="target directory")
    parser.add_argument(
        "-i",
        dest="inifile",
        action="store",
        type=str,
        default="idefix.ini",
        help="target inifile",
    )
    time_group = parser.add_mutually_exclusive_group()
    time_group.add_argument(
        "--duration",
        action="store",
        type=float,
        help="run for specified time (in code units)",
    )
    time_group.add_argument(
        "--one",
        "--one-step",
        dest="one_step",
        action="store_true",
        help="run only for one time step",
    )
    parser.add_argument(
        "--time-step",
        dest="time_step",
        action="store",
        type=float,
        help="patch the inifile TimeIntegrator.first_dt parameter",
    )


def command(
    directory: str,
    inifile: str = "idefix.ini",
    duration: float | None = None,
    time_step: float | None = None,
    one_step: bool | None = False,
) -> int:

    input_inifile = inifile
    for loc in [Path.cwd(), Path(directory)]:
        pinifile = (loc / input_inifile).resolve()
        if pinifile.is_file():
            break
    else:
        print_err(f"could not find inifile {input_inifile}")
        return 1
    if one_step:
        if time_step is None:
            time_step = inifix.load(pinifile)["TimeIntegrator"]["first_dt"]
        duration = time_step

    d = Path(directory)
    if not (d / "idefix").is_file():
        if not (d / "Makefile").is_file():
            print_err(
                "No idefix instance or Makefile found in the target directory. "
                "Run `idfx setup` first."
            )
            return 1

        if (ret := _make(directory)) != 0:
            return ret

    conf = inifix.load(pinifile)
    if time_step is not None:
        conf["TimeIntegrator"]["first_dt"] = time_step
    if duration is not None:
        conf["TimeIntegrator"]["tstop"] = duration

    with pushd(d), NamedTemporaryFile() as tmp_inifile:
        conf.write(tmp_inifile.name)
        ret = call(["./idefix", "-i", tmp_inifile.name])
        if ret != 0:
            print_err("idefix terminated with an error.")
    return ret



"""write an Idefix inifile a from json string"""
from __future__ import annotations

import argparse
import json
import sys
from pathlib import Path

import inifix

from idefix_cli._commons import print_err


def add_arguments(parser) -> None:
    parser.add_argument("dest", type=str, help="dest inifile")
    parser.add_argument(
        "source",
        nargs="?",
        type=argparse.FileType("r"),
        default=sys.stdin,
        help="json input",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="force writing even if savefile already exists",
    )


def command(dest: str, source, force: bool = False) -> int:
    try:
        data = json.load(source)
    except json.decoder.JSONDecodeError:
        print_err("input is not valid json.")
        return 1

    try:
        inifix.validate_inifile_schema(data)
    except ValueError:
        print_err("input is not Pluto inifile format compliant.")
        return 1

    dest = Path(dest)
    if dest.is_file() and not force:
        print_err(
            f"destination file {dest} already exists. Use -f/--force to overwrite."
        )
        return 1

    inifix.dump(data, dest)
    return 0



"""print relevant data for reproduction to stdout"""
from __future__ import annotations

import json

from idefix_cli._commons import get_git_data
from idefix_cli._commons import requires_idefix


def add_arguments(parser) -> None:
    parser.add_argument(
        "-d",
        "--json",
        dest="todict",
        action="store_true",
        help="print to JSON serializable format.",
    )


@requires_idefix()
def command(todict: bool = False) -> int:
    """
    Print idefix latest version tag-(git hash) and current time to stdout.
    """
    data = get_git_data()
    if todict:
        text = json.dumps(data, indent=2)
    else:
        text = "\n".join(list(data.values()))

    print(text)
    return 0



"""read an Idefix inifile and print it to json format"""
from __future__ import annotations

import json
from pathlib import Path

from inifix import load

from idefix_cli._commons import print_err


def add_arguments(parser) -> None:
    parser.add_argument("inifile", type=str, help="target inifile")
    parser.add_argument(
        "--indent", type=int, help="indentation in spaces (default is flat output)"
    )


def command(inifile: str, indent: int | None = None) -> int:
    inifile = Path(inifile)
    if not inifile.is_file():
        print_err(f"no such file {inifile}")
        return 1
    print(json.dumps(load(inifile), indent=indent))
    return 0



from __future__ import annotations

import argparse
import sys
from importlib import import_module
from pkgutil import walk_packages
from types import FunctionType
from typing import cast
from typing import NoReturn

from idefix_cli import __version__
from idefix_cli import _commands
from idefix_cli._commons import print_err


def _setup_commands(parser: argparse.ArgumentParser) -> dict[str, FunctionType]:
    # https://github.com/python/mypy/issues/1422
    path: str = _commands.__path__  # type: ignore

    sparsers = parser.add_subparsers(dest="command")
    retv: dict[str, FunctionType] = {}
    for module_info in walk_packages(path, _commands.__name__ + "."):
        module = import_module(module_info.name)

        _, _, command_name = module_info.name.rpartition(".")

        # plugin validation
        # TODO: could validate signatures too
        if not hasattr(module, "add_arguments"):
            raise RuntimeError(
                f"command plugin {command_name} is missing an 'add_arguments' function."
            )
        if not hasattr(module, "command"):
            raise RuntimeError(
                f"command plugin {command_name} is missing a 'command' function."
            )

        # optional plugin hooks
        kwargs = getattr(module, "parser_kwargs", {})

        sub_parser = sparsers.add_parser(command_name, help=module.__doc__, **kwargs)
        module.add_arguments(sub_parser)  # type: ignore
        retv.update({command_name: module.command})  # type: ignore
    return retv


def main(argv: list[str] | None = None) -> int | NoReturn:
    parser = argparse.ArgumentParser(prog="idfx")
    parser.add_argument("-v", "--version", action="version", version=__version__)
    commands = _setup_commands(parser)

    known_args, unknown_args = parser.parse_known_args(argv)

    vargs = vars(known_args)
    cmd_name = vargs.pop("command")

    if cmd_name is None:
        # calling `idfx` without any argument is equivalent to `idfx --help`
        # except that the return value non zero.
        parser.print_help(sys.stderr)
        return 1

    cmd = commands[cmd_name]
    if cmd_name == "conf":
        return cmd(*unknown_args, **vars(known_args))  # type: ignore

    elif unknown_args:
        print_err(f"received unknown arguments {tuple(unknown_args)!r}.")
        return 1

    retv = cmd(**vargs)
    retv = cast(int, retv)
    return retv
